﻿namespace StockSharp.Plaza
{
	using System;
	using System.Collections.Generic;
	using System.Diagnostics;
	using System.Linq;

	using Ecng.Collections;

	using StockSharp.BusinessEntities;
	using StockSharp.Plaza.Metadata;

	class PlazaDepthBuilder
	{
        private sealed class QuoteInfo
        {
            public QuoteInfo(long replRev, Quote quote)
            {
                ReplRev = replRev;
                Quote = quote;
            }

            public long ReplRev { get; private set; }
            public Quote Quote { get; private set; }
        }

		private sealed class DepthInfo
		{
			public DepthInfo(MarketDepth pendingDepth)
			{
				if (pendingDepth == null)
					throw new ArgumentNullException("pendingDepth");

				PendingDepth = pendingDepth;
                Quotes = new SynchronizedDictionary<long, QuoteInfo>();
			}

            public SynchronizedDictionary<long, QuoteInfo> Quotes { get; private set; }
			public MarketDepth PendingDepth { get; private set; }
		}

		private readonly PlazaTrader _trader;

		private readonly SynchronizedDictionary<PlazaStream, SynchronizedDictionary<Security, DepthInfo>> _info =
			new SynchronizedDictionary<PlazaStream, SynchronizedDictionary<Security, DepthInfo>>();

		private readonly SynchronizedList<MarketDepth> _changedDepths = new SynchronizedList<MarketDepth>();

		public PlazaDepthBuilder(PlazaTrader trader)
		{
			if (trader == null)
				throw new ArgumentNullException("trader");

			_trader = trader;

			_info.SafeAdd(trader.StreamRegistry.Aggregation5FutureStream);
			_info.SafeAdd(trader.StreamRegistry.Aggregation20FutureStream);
			_info.SafeAdd(trader.StreamRegistry.Aggregation50FutureStream);
			_info.SafeAdd(trader.StreamRegistry.Aggregation5OptionStream);
			_info.SafeAdd(trader.StreamRegistry.Aggregation20OptionStream);
			_info.SafeAdd(trader.StreamRegistry.Aggregation50OptionStream);
		}

		public void ClearQuotes(PlazaStream stream)
		{
			_info[stream].Clear();
		}

		public void UpdateQuote(PlazaStream stream, Security security, PlazaRecord record, PlazaAggregationColumns metadata)
		{
			var quote = new Quote
			{
				Security = security,
				Price = record.Get<decimal>(metadata.Price),
				Volume = record.Get<int>(metadata.Volume),
				OrderDirection = record.Get<int>(metadata.Direction) == 1 ? OrderDirections.Buy : OrderDirections.Sell,
				ExtensionInfo = new Dictionary<object, object> { { "replID", record.ReplId } }
			};

			Debug.WriteLine("Q " + quote.Security.Id + "," + record.ReplId + "," + record.ReplAct + "," + record.ReplRev + "," + quote.Price + "@" + quote.Volume + "," + quote.OrderDirection);

			var oldDepth = _trader.GetMarketDepth(security);

			var info = _info[stream].SafeAdd(security, key => new DepthInfo(oldDepth.Clone()));

			//bool skip = false;
			var quotes = info.Quotes.SyncGet(d =>
			{
				if (record.ReplAct == 0)
				{
					// Судя по этой ветке http://forum.rts.ru/viewtopic.asp?t=16922
					// В безбазовом клиенте запись нужно удалять, если и объем, и цена равны нулю.
					if (quote.Volume != 0 && quote.Price != 0)
					{
						//	var old = d.TryGetValue(record.ReplId);
						//	if (old != null && old.Equals(quote))
						//		skip = true;
						//	else
						d[record.ReplId] = new QuoteInfo(record.ReplRev, quote);
					}
					else if (d.ContainsKey(record.ReplId))
						d.Remove(record.ReplId);
					//	else
					//		skip = true;
				}
				else
				{
					// P2ClientFate.doc - стр. 16
					// replAct — признак того, что запись удалена. При удалении записи на сервере в поле replAct заносится значение ее replID. 
					// Если replAct = 0 — запись активна (не удалена)
					// Встречается только в БАЗОВОМ клиенте
					if (d.ContainsKey(record.ReplId))
						d.Remove(record.ReplId);
					//	else
					//		skip = true;
				}

                return ConstructQuoteMap(d.Values).Values.ToArray();
			});

			var time = record.Get<DateTime>(metadata.Moment);

			var oldTime = oldDepth.LastChangeTime;

			info.PendingDepth.Update(quotes);	// неявно ставит time=DateTime.Now

			//_pendingDepths[security] = newDepth;
			
			// В соответствии с http://forum.rts.ru/viewtopic.asp?t=16922
			// у следующей записи может быть moment < moment предыдущей записи
			// например когда обновляется объем у старого "уровня" в стакане.
			// т.о. в сценарии когда в стакане обновляются уровни, а не добавляются новые
			// время последнего обновления получить из только этого потока репликации невозможно.
			// в качестве временной меры за время стакана берется максимальное значение moment 
			// для данного инструмента/

			info.PendingDepth.LastChangeTime = (time - oldTime) >= TimeSpan.FromMilliseconds(1) ? time : oldTime.AddMilliseconds(1);

			_changedDepths.Add(oldDepth);

			//SimpleDepthLogger.Append(marketDepth.Security.Id.Replace('@', '_') + ".lgg",
			//	new[] { marketDepth });
		}

        public void RemoveQuotesByRev(PlazaStream stream, long revision)
        {
            if (!_info.ContainsKey(stream))
                return;

            var dictInfo = _info[stream];

            foreach (var info in dictInfo)
            {
                var security = info.Key;
                var oldDepth = _trader.GetMarketDepth(security);
                var depthInfo = dictInfo.SafeAdd(security, key => new DepthInfo(oldDepth.Clone()));

                var quotes = depthInfo.Quotes.SyncGet(d =>
                {
                    d.RemoveWhere(qs => qs.Value.ReplRev < revision);

                    return ConstructQuoteMap(d.Values).Values.ToArray();
                });

                var time = _trader.MarketTime;

                var oldTime = oldDepth.LastChangeTime;

                depthInfo.PendingDepth.Update(quotes);	// неявно ставит time=DateTime.Now
                depthInfo.PendingDepth.LastChangeTime = (time - oldTime) >= TimeSpan.FromMilliseconds(1) ? time : oldTime.AddMilliseconds(1);

                _changedDepths.Add(oldDepth);
            }
        }

        private Dictionary<decimal, Quote> ConstructQuoteMap(IEnumerable<QuoteInfo> quoteInfos)
        {
            // http://forum.rts.ru/viewtopic.asp?p=101663. 
            // Может быть одна строка стакана (одна цена) но два разных replId. 
            // Они дублируют друг друга и возникают когда заявка входит-выходит 
            // за границу в 5 котировок
            // Поэтому здесь агрегируем по цене. При этом иногда возникают ситуации что
            // после обработки события дефакто стакан не изменился. Они отфильтровываются ниже (MD.GetDelta)

            var quoteMap = new Dictionary<decimal, Quote>();

            foreach (var q in quoteInfos)
            {
                quoteMap[q.Quote.Price] = q.Quote;

                if (q.Quote.Volume == 0)
                    throw new InvalidOperationException("quote Volume=0");
            }

            return quoteMap;
        }

		public IEnumerable<MarketDepth> FlushChanges(PlazaStream stream)
		{
			var dict = _info[stream];

			return _changedDepths.SyncGet(c =>
			{
				foreach (var oldDepth in c)
				{
					var newDepth = dict[oldDepth.Security].PendingDepth;

					// NOTE: для ускорения пусть это будет проверятся на стороне клиента
					//var delta = oldDepth.GetDelta(newDepth);

					//if (delta.Bids.Length == 0 && delta.Asks.Length == 0)
					//{
					//    // ничего не изменилось. 
					//    newDepth.LastChangeTime = oldDepth.LastChangeTime;
					//}
					//else
					//{
					//Debug.WriteLine("PLAZA:" + SimpleDepthLogger.ToString(newDepth));

					if (newDepth.Bids.Length > 0 && newDepth.Asks.Length > 0 && newDepth.BestBid.Price >= newDepth.BestAsk.Price)
						throw new InvalidOperationException("BestBid>=BestAsk");

					oldDepth.Security.BestBid = newDepth.BestBid;
					oldDepth.Security.BestAsk = newDepth.BestAsk;


					oldDepth.Update(newDepth.Bids, newDepth.Asks, true);
					oldDepth.LastChangeTime = newDepth.LastChangeTime;
					//}
				}

				return c.CopyAndClear();
			});
		}
	}
}