﻿namespace StockSharp.Plaza
{
	using System;
	using System.Collections.Generic;
	using System.Linq;

	using Ecng.Common;

	using StockSharp.Plaza.Metadata;
	using StockSharp.Plaza.Wrappers;

	/// <summary>
	/// Запись в потоке репликации.
	/// </summary>
	public class PlazaRecord : Cloneable<PlazaRecord>
	{
		private readonly int _replIdIndex;
		private readonly int _replRevIndex;
		private readonly int _replActIndex;

		internal PlazaRecord(PlazaTable table)
		{
			if (table == null)
				throw new ArgumentNullException("table");

			Table = table;

			_replIdIndex = 0;// Table.Columns.IndexOf(c => c.Name == "replID");
			_replRevIndex = 1;// Table.Columns.IndexOf(c => c.Name == "replRev");
			_replActIndex = 2;// Table.Columns.IndexOf(c => c.Name == "replAct");
		}

        private object[] _data;
        
        private object[] Data
        {
			get { return _data ?? (_data = new object[Table.Columns.Count]); }
        	set
            {
                if (value == null)
                    throw new ArgumentNullException("value");

				_data = value;
            }
        }

		/// <summary>
		/// Таблица с метаданными, описывающая данные этой записи.
		/// </summary>
		public PlazaTable Table { get; private set; }

		/// <summary>
		/// Количество колонок в записи.
		/// </summary>
		public int Count
		{
			get { return Data.Length; }
		}

		/// <summary>
		/// Перечислитель, содержащий значения колонок записи.
		/// </summary>
		public IEnumerable<object> Values
		{
			get { return Data; }
		}

		/// <summary>
		/// Получить значение из записи по колонке.
		/// </summary>
		/// <typeparam name="T">Тип возвращаемого значения.</typeparam>
		/// <param name="column">Колонка.</param>
		/// <returns>Значение.</returns>
		public T Get<T>(PlazaColumn column)
		{
            var item = Table.Columns.IndexOf(column);
            if (item == -1)
                throw new ArgumentException("Колонка {0} не присутствует в потоке.".Put(column.Name));

			var value = Data[item];

			try
			{
				if (typeof(T) == typeof(DateTime))
					return value.ToString().ToDateTime().To<T>();

				if (value is int && typeof(T) == typeof(string))
					return value.ToString().To<T>();

				if (value is int && typeof(T) == typeof(bool))
					return ((int)value == 1).To<T>();

				if (value is sbyte && typeof(T) == typeof(bool))
					return ((sbyte)value == 1).To<T>();

				if (value is long && typeof(T) == typeof(int))
					return ((long)value).To<T>();

				if (value is sbyte && typeof(T) == typeof(int))
					return ((sbyte)value).To<T>();

				return value.To<T>();
			}
			catch (Exception ex)
			{
				throw new InvalidCastException("Колонка {0} таблицы {1} имеет тип '{2}'. Приведение к типу '{3}' невозможно."
					.Put(column.Name, column.TableId, value == null ? null : value.GetType(), typeof(T)), ex);
			}
		}

		/// <summary>
		/// Уникальный идентификатор записи в таблице (обязательная колонка).
		/// </summary>
		public long ReplId
		{
			get { return (long)Data[_replIdIndex]; }
		}

		/// <summary>
		/// Уникальный номер изменения в таблице (обязательная колонка).
		/// </summary>
		public long ReplRev
		{
			get { return (long)Data[_replRevIndex]; }
		}

		/// <summary>
		/// Признак того, что запись удалена (обязательная колонка).
		/// Если replAct = 0 — запись активна (не удалена).
		/// </summary>
		public long ReplAct
		{
			get { return (long)Data[_replActIndex]; }
		}

		internal PlazaRecord Fill(IRecordWrapper record)
		{
			if (record == null)
				throw new ArgumentNullException("record");

			for (var i = 0u; i < Data.Length; i++)
			{
				Data[i] = record.GetObject(i);
			}

			return this;
		}

		internal void Set(PlazaColumn column, object value)
		{
			if (column == null)
				throw new ArgumentNullException("column");

			Data[Table.Columns.IndexOf(column)] = value;
		}

		/// <summary>
		/// Создать копию.
		/// </summary>
		/// <returns>Копия.</returns>
		public override PlazaRecord Clone()
		{
			return new PlazaRecord(Table)
			{
				Data = Data.ToArray()
			};
		}
	}
}