﻿namespace StockSharp.Plaza
{
	using System;
	using System.IO;
	using System.Runtime.InteropServices;

	using Ecng.Collections;
	using Ecng.Common;

	using StockSharp.Plaza.Metadata;
	using StockSharp.Plaza.Wrappers;

	using Ecng.Interop;

	class PlazaStreamManager : Disposable
	{
        private readonly TimeSpan _errorSleepInterval = TimeSpan.FromMilliseconds(2000);
		private readonly PlazaTableSerializer _configSerializer = new PlazaTableSerializer();

		private readonly PlazaConnectionPool _connectionPool;
		private readonly Action<Exception> _errorHandler;

		private PlazaConnectionToken _token;

		public PlazaStreamManager(PlazaConnectionPool connectionPool, Action<Exception> errorHandler)
		{
			if (connectionPool == null)
				throw new ArgumentNullException("connectionPool");

			if (errorHandler == null)
				throw new ArgumentNullException("errorHandler");

			_connectionPool = connectionPool;
			_errorHandler = errorHandler;

			Streams = new PlazaStreamList(this);
		}

		private string _configPath;

		public string ConfigPath
		{
			get { return _configPath; }
			set
			{
				_configPath = value;
				Streams.ForEach(s => s.ConfigFileName = string.Empty);
				ConfigPath.CreateDirIfNotExists();
			}
		}

		public PlazaStreamList Streams { get; private set; }

		public bool IsExportStarted
		{
			get { return _token != null; }
		}

		public void StartExport()
		{
			if (Streams.Count == 0)
				throw new InvalidOperationException("Коллекция потоков данных пуста.");

			foreach (var stream in Streams)
			{
				if (stream.ConfigFileName.IsEmpty())
				{
					string config;

					if (ConfigPath.IsEmpty())
					{
						var tempFileName = Path.GetTempFileName();
						config = tempFileName.Replace("tmp", "ini");
						File.Delete(tempFileName);
						_configSerializer.Serialize(stream.Table, config);
					}
					else
					{
						config = Path.Combine(ConfigPath, stream.Table.Id.ToString().Replace("\\", "_") + ".ini");

						if (!File.Exists(config))
							_configSerializer.Serialize(stream.Table, config);
					}

					stream.ConfigFileName = config;
				}
			}

			_token = _connectionPool.RunConnection("Plaza. Stream thread.", connection =>
			{
				if (connection.Status != (ConnectionWrapperStatus.Connected | ConnectionWrapperStatus.RouterConnected))
				{
					CloseStreams();
					_errorSleepInterval.Sleep();
					return true;
				}

				PlazaException lastError = null;

				Streams.SyncDo(c => c.ForEach(stream =>
				{
					try
					{
						stream.ProcessConnection(connection);
					}
					catch (COMException e)
					{
						//System.Diagnostics.Trace.WriteLine("Ошибка проверки потока репликации - stream.CheckConnection(_connection) - COMException " + e.ErrorCode.ToString());
						lastError = new PlazaException("Ошибка обработки потока данных.", e);
					}
				}));

				if (lastError != null)
				{
					_errorHandler(lastError);
					_errorSleepInterval.Sleep();
				}

				return true;
			},
			connection => CloseStreams());
		}

		public void StopExport()
		{
			if (!IsExportStarted)
				return;

			_connectionPool.StopConnection(_token);
			_token = null;
		}

		private void CloseStreams()
		{
			Streams.ForEach(s => s.Close());
		}

		protected override void DisposeManaged()
		{
			StopExport();
			Streams.Clear();
			base.DisposeManaged();
		}
	}
}
