﻿namespace StockSharp.Plaza
{
	using System;
	using System.Collections.Generic;
	using System.IO;
	using System.Linq;
	using System.Net;
	using System.Text;

	using Ecng.Common;
	using Ecng.Collections;

	using StockSharp.Algo.Logging;
	using StockSharp.Algo.Storages;
	using StockSharp.BusinessEntities;
	using StockSharp.Algo;
	using StockSharp.Plaza.Metadata;
	using StockSharp.Plaza.Properties;

	/// <summary>
	/// Реализация интерфейса <see cref="ITrader"/> для взаимодействия со шлюзом РТС через Plaza 2.
	/// </summary>
	public class PlazaTrader : BaseTrader
	{
		private readonly SynchronizedSet<int> _isinIds = new SynchronizedSet<int>();
		private readonly SynchronizedSet<Security> _multilegSecurities = new SynchronizedSet<Security>();

		private readonly PlazaStreamManager _streamManager;

		private const string _futuresClass = "FUT";
		private const string _optionsClass = "OPT";

		private const string _iniFile = "P2ClientGate.ini";

		private readonly PlazaConnectionPool _connectionPool;

		private PlazaDepthBuilder _depthBuilder;
		private PlazaOrderBuilder _orderBuilder;

		private PlazaStream[] _defaultStreams;

		/// <summary>
		/// Создать <see cref="PlazaTrader"/>.
		/// </summary>
		public PlazaTrader()
			: this(new IPEndPoint(IPAddress.Loopback, PlazaAddresses.DefaultPlaza2Port))
		{
		}

		/// <summary>
		/// Создать <see cref="PlazaTrader"/>.
		/// </summary>
		/// <param name="address">Адрес сервера.</param>
		public PlazaTrader(EndPoint address)
		{
			if (!File.Exists(_iniFile))
				File.WriteAllText(_iniFile, Resources.P2ClientGate, Encoding.UTF8);

			if (!File.Exists(TransactionFactory.MessagesFileName))
				File.WriteAllText(TransactionFactory.MessagesFileName, Resources.p2fortsgate_messages, Encoding.UTF8);

			DefaultSecurityClassInfo = new RefPair<SecurityTypes, Exchange>(SecurityTypes.Future, Exchange.Rts);

			_connectionPool = new PlazaConnectionPool(RaiseConnected, RaiseDisconnected, RaiseConnectionError, RaiseProcessDataError, (type, msg) => RaiseLog(type, () => msg)) { Address = address };
			_streamManager = new PlazaStreamManager(_connectionPool, RaiseProcessDataError);
			StreamRegistry = new PlazaStreamRegistry();

			_transactionManager = new TransactionManager(_connectionPool, TransactionIdGenerator);
			_transactionManager.ProcessResponse += ProcessReply;

			OnlySystemTrades = true;
		}

		private new void RaiseConnectionError(Exception exception)
		{
			// РТС Техподдержка: "при обрыве связи нужно вызвать метод Disconnect"
			_connectionPool.Disconnect();

			this.AddInfoLog("RaiseConnectionError - disconnected");

			base.RaiseConnectionError(exception);
		}

		private void OnStreamLog(PlazaStream stream, string message)
		{
			RaiseLog(() => "PlazaStream {0}:{1}".Put(stream.Table.Id, message));
		}

		/// <summary>
		/// Логин.
		/// </summary>
		public string Login
		{
			get { return _connectionPool.Login; }
			set { _connectionPool.Login = value; }
		}

		/// <summary>
		/// Пароль.
		/// </summary>
		public string Password
		{
			get { return _connectionPool.Password; }
			set { _connectionPool.Password = value; }
		}

		/// <summary>
		/// Адрес сервера. По-умолчанию значение равно адресу локального компьютера 127.0.0.1:4001.
		/// </summary>
		public EndPoint Address
		{
			get { return _connectionPool.Address; }
			set { _connectionPool.Address = value; }
		}

		/// <summary>
		/// Префикс в имени приложения Plaza соединения.
		/// </summary>
		/// <remarks>
		/// Все созданные <see cref="PlazaTrader"/> соединения именуются как <see cref="AppName"/>_N, где N - порядковый номер соединение в пределах одного робота.
		/// Если планируется подключать несколько <see cref="PlazaTrader"/> к одному роутеру, они должны иметь уникальные <see cref="AppName"/>.
		/// </remarks>
		public string AppName
		{
			get { return _connectionPool.AppName; }
			set { _connectionPool.AppName = value; }
		}

		/// <summary>
		/// Время, в течение которого ожидается получение сообщения из потока данных.
		/// </summary>
		/// <remarks>
		/// По-умолчанию равно 100 милисекундам.
		/// </remarks>
		public TimeSpan StreamTimeOut
		{
			get { return _connectionPool.PollTimeOut; }
			set { _connectionPool.PollTimeOut = value; }
		}

		/// <summary>
		/// Путь к директории, в которой будут храниться схемы потоков Plaza.
		/// </summary>
		/// <remarks>
		/// По-умолчанию путь не инициализирован. Это означает, что схемы будут генерироваться каждый раз во временном хранилище.
		/// </remarks>
		public string StreamConfigPath
		{
			get { return _streamManager.ConfigPath; }
			set { _streamManager.ConfigPath = value; }
		}

		/// <summary>
		/// Использовать SharedMem протокол при соединении с локальным роутером. По умолчанию false.
		/// </summary>
		public bool UseLocalProtocol
		{
			get { return _connectionPool.UseLocalProtocol; }
			set { _connectionPool.UseLocalProtocol = value; }
		}

		/// <summary>
		/// Использовать полный ордер лог. По умолчанию false;
		/// </summary>
		public bool UseFullOrdersLog { get; set; }

		private PlazaStreamRegistry _streamRegistry;

		/// <summary>
		/// Реестр всех потоков данных сервера Плазы.
		/// </summary>
		public PlazaStreamRegistry StreamRegistry
		{
			get { return _streamRegistry; }
			set
			{
				if (value == null)
					throw new ArgumentException("value");

				if (IsExportStarted)
					throw new InvalidOperationException("Запущена передача данных. Смена потоков данных сервера Плазы невозможна.");

				if (_streamRegistry != value)
				{
					if (_streamRegistry != null)
					{
						UnSubscribeStreamRegistryEvents(_streamRegistry);

						if (_streamManager != null)
							_streamManager.Streams.Clear();

						_defaultStreams.Clear();
					}

					_streamRegistry = value;
					SubscribeStreamRegistryEvents(_streamRegistry);

					DefaultFutureDepthStream = _streamRegistry.Aggregation5FutureStream;
					DefaultOptionDepthStream = _streamRegistry.Aggregation5OptionStream;

					_defaultStreams = new[] {
						_streamRegistry.CommonFutureStream,
						_streamRegistry.CommonOptionStream,
						_streamRegistry.SessionContentsOptionStream,
						_streamRegistry.SessionContentsFutureStream,
						_streamRegistry.PortfoliosStream,
						_streamRegistry.PositionsStream,
						_streamRegistry.DealFutureStream,
						_streamRegistry.DealOptionStream,
						_streamRegistry.OrdersLogFutureStream,
						_streamRegistry.OrdersLogOptionStream,
						_streamRegistry.SessionFutureStream,
						// по-умолчанию поток отключен
						//_streamRegistry.VolatilityStream,
						DefaultFutureDepthStream,
						DefaultOptionDepthStream
					};

					if (_streamManager != null)
					{
						_streamManager.Streams.SyncDo(streams => streams.AddRange(_defaultStreams.Where(stream => !streams.Contains(stream))));
					}

					_orderBuilder = new PlazaOrderBuilder(this);
					_depthBuilder = new PlazaDepthBuilder(this);
				}
			}
		}

		private SessionState _sessionState;

		/// <summary>
		/// Состояние торговой сессии.
		/// </summary>
		public SessionState SessionState
		{
			get { return _sessionState; }
			private set
			{
				if (_sessionState != value)
				{
					_sessionState = value;
					SessionStateChanged.SafeInvoke();
				}
			}
		}

		/// <summary>
		/// Событие изменения <see cref="SessionState"/>.
		/// </summary>
		public event Action SessionStateChanged;

		private readonly object _routerLock = new object();
		private PlazaRouter _router;

		/// <summary>
		/// Получить управляение и настройки программы P2MQRouter.exe, которая осуществляет подключение к Plaza 2.
		/// </summary>
		public PlazaRouter Router
		{
			get
			{
				lock (_routerLock)
					return _router ?? (_router = new PlazaRouter(_connectionPool));
			}
		}

		private readonly TransactionManager _transactionManager;

		/// <summary>
		/// Менеджер транзакций.
		/// </summary>
		public TransactionManager TransactionManager
		{
			get { return _transactionManager; }
		}

		/// <summary>
		/// Передавать через событие <see cref="ITrader.NewTrades"/> только системные сделки. По-умолчанию значение равно true.
		/// </summary>
		public bool OnlySystemTrades { get; set; }

		private PlazaStream _defaultFutureDepthStream;

		/// <summary>
		/// Поток данных для стакана фьючерсных контрактов, который будет запускаться при вызове метода <see cref="StartExport"/>.
		/// </summary>
		/// <remarks>
		/// Значение по умолчанию равно <see cref="PlazaStreamRegistry.Aggregation5FutureStream"/>.
		/// </remarks>
		public PlazaStream DefaultFutureDepthStream
		{
			get { return _defaultFutureDepthStream; }
			set
			{
				if (value == null)
					throw new ArgumentNullException("value");

				if (IsExportStarted)
					throw new InvalidOperationException("Запущена передача данных. Смена дефолтного потока данных для стакана фьючерсных контрактов невозможна.");

				if (!StreamRegistry.AggregationFutureStreams.Contains(value))
					throw new ArgumentException("{0} не является потоком агрегированных стаканов фьючерсов.".Put(value.Table.Id));

				if (_defaultFutureDepthStream != value)
				{
					if (_defaultFutureDepthStream != null)
					{
						UnSubscribeDepthStream(_defaultFutureDepthStream);

						Streams.SyncDo(c =>
						{
							if (c.Contains(_defaultFutureDepthStream))
								c.Remove(_defaultFutureDepthStream);

							c.Add(value);
						});
					}

					_defaultFutureDepthStream = value;

					SubscribeDepthStream(_defaultFutureDepthStream);
				}
			}
		}

		private PlazaStream _defaultOptionDepthStream;

		/// <summary>
		/// Поток данных для стакана опционных контрактов, который будет запускаться при вызове метода <see cref="StartExport"/>.
		/// </summary>
		/// <remarks>
		/// Значение по умолчанию равно <see cref="PlazaStreamRegistry.Aggregation5OptionStream"/>.
		/// </remarks>
		public PlazaStream DefaultOptionDepthStream
		{
			get { return _defaultOptionDepthStream; }
			set
			{
				if (value == null)
					throw new ArgumentNullException("value");

				if (IsExportStarted)
					throw new InvalidOperationException("Запущена передача данных. Смена дефолтного потока данных для стакана опционов невозможна.");

				if (!StreamRegistry.AggregationOptionStreams.Contains(value))
					throw new ArgumentException("{0} не является потоком агрегированных стаканов опционов.".Put(value.Table.Id));

				if (_defaultOptionDepthStream != value)
				{
					if (_defaultOptionDepthStream != null)
					{
						UnSubscribeDepthStream(_defaultOptionDepthStream);

						Streams.SyncDo(c =>
						{
							if (c.Contains(_defaultOptionDepthStream))
								c.Remove(_defaultOptionDepthStream);

							c.Add(value);
						});
					}

					_defaultOptionDepthStream = value;
					
					SubscribeDepthStream(_defaultOptionDepthStream);
				}
			}
		}

		/// <summary>
		/// Проверить соединение.
		/// </summary>
		public override bool IsConnected
		{
			get { return _connectionPool != null && _connectionPool.IsConnected; }
		}

		/// <summary>
		/// Запущен ли экспорт. Экспорт запускается через метод <see cref="M:StockSharp.Algo.BaseTrader.StartExport"/>.
		/// </summary>
		public override bool IsExportStarted
		{
			get { return _streamManager != null && _streamManager.IsExportStarted; }
		}

		/// <summary>
		/// Потоки, которые будут запущены при вызове метода <see cref="StartExport"/>.
		/// </summary>
		public PlazaStreamList Streams
		{
			get { return _streamManager.Streams; }
		}

		private void SubscribeStreamRegistryEvents(PlazaStreamRegistry streamRegistry)
		{
			streamRegistry.CommonFutureStream.Inserted += OnCommonFutureStreamInserted;
			streamRegistry.SessionContentsFutureStream.Inserted += OnSessionContentsFutureStreamInserted;
			streamRegistry.CommonOptionStream.Inserted += OnCommonOptionStreamInserted;
			streamRegistry.SessionContentsOptionStream.Inserted += OnSessionContentsOptionStreamInserted;
			streamRegistry.PortfoliosStream.Inserted += OnPortfoliosStreamInserted;
			streamRegistry.PositionsStream.Inserted += OnPositionsStreamInserted;

			streamRegistry.Aggregation5OptionStream.Inserted += OnAggregation5OptionStreamInserted;
			streamRegistry.Aggregation5FutureStream.Inserted += OnAggregation5FutureStreamInserted;
			streamRegistry.Aggregation20OptionStream.Inserted += OnAggregation20OptionStreamInserted;
			streamRegistry.Aggregation20FutureStream.Inserted += OnAggregation20FutureStreamInserted;
			streamRegistry.Aggregation50OptionStream.Inserted += OnAggregation50OptionStreamInserted;
			streamRegistry.Aggregation50FutureStream.Inserted += OnAggregation50FutureStreamInserted;
			streamRegistry.SessionFutureStream.Inserted += OnSessionFutureStreamInserted;
			streamRegistry.DealFutureStream.Inserted += OnDealFutureStreamInserted;
			streamRegistry.DealOptionStream.Inserted += OnDealOptionStreamInserted;
			streamRegistry.OrdersLogFutureStream.Inserted += OnOrdersLogFutureStreamInserted;
			streamRegistry.OrdersLogOptionStream.Inserted += OnOrdersLogOptionStreamInserted;

			streamRegistry.OrdersLogFutureStream.End += OnOrdersLogStreamEnd;
			streamRegistry.OrdersLogOptionStream.End += OnOrdersLogStreamEnd;

			streamRegistry.OrdersLogFutureStream.Flushed += OnOrdersLogStreamFlushed;
			streamRegistry.OrdersLogOptionStream.Flushed += OnOrdersLogStreamFlushed;

			streamRegistry.VolatilityStream.Inserted += OnVolatilityStreamInserted;

			streamRegistry.AllStreams.ForEach(s => s.Log += OnStreamLog);
		}

		private void UnSubscribeStreamRegistryEvents(PlazaStreamRegistry streamRegistry)
		{
			streamRegistry.CommonFutureStream.Inserted -= OnCommonFutureStreamInserted;
			streamRegistry.SessionContentsFutureStream.Inserted -= OnSessionContentsFutureStreamInserted;
			streamRegistry.CommonOptionStream.Inserted -= OnCommonOptionStreamInserted;
			streamRegistry.SessionContentsOptionStream.Inserted -= OnSessionContentsOptionStreamInserted;
			streamRegistry.PortfoliosStream.Inserted -= OnPortfoliosStreamInserted;
			streamRegistry.PositionsStream.Inserted -= OnPositionsStreamInserted;

			streamRegistry.Aggregation5OptionStream.Inserted -= OnAggregation5OptionStreamInserted;
			streamRegistry.Aggregation5FutureStream.Inserted -= OnAggregation5FutureStreamInserted;
			streamRegistry.Aggregation20OptionStream.Inserted -= OnAggregation20OptionStreamInserted;
			streamRegistry.Aggregation20FutureStream.Inserted -= OnAggregation20FutureStreamInserted;
			streamRegistry.Aggregation50OptionStream.Inserted -= OnAggregation50OptionStreamInserted;
			streamRegistry.Aggregation50FutureStream.Inserted -= OnAggregation50FutureStreamInserted;
			streamRegistry.SessionFutureStream.Inserted -= OnSessionFutureStreamInserted;
			streamRegistry.DealFutureStream.Inserted -= OnDealFutureStreamInserted;
			streamRegistry.DealOptionStream.Inserted -= OnDealOptionStreamInserted;
			streamRegistry.OrdersLogFutureStream.Inserted -= OnOrdersLogFutureStreamInserted;
			streamRegistry.OrdersLogOptionStream.Inserted -= OnOrdersLogOptionStreamInserted;

			streamRegistry.OrdersLogFutureStream.End -= OnOrdersLogStreamEnd;
			streamRegistry.OrdersLogOptionStream.End -= OnOrdersLogStreamEnd;

			streamRegistry.OrdersLogFutureStream.Flushed -= OnOrdersLogStreamFlushed;
			streamRegistry.OrdersLogOptionStream.Flushed -= OnOrdersLogStreamFlushed;

			streamRegistry.VolatilityStream.Inserted -= OnVolatilityStreamInserted;

			streamRegistry.AllStreams.ForEach(s => s.Log -= OnStreamLog);
		}

		private void SubscribeDepthStream(PlazaStream depthStream)
		{
			depthStream.End += OnAggregationStreamEnd;
			depthStream.Flushed += OnAggregationStreamFlushed;
			depthStream.Cleared += OnAggregationStreamCleared;
		}

		private void UnSubscribeDepthStream(PlazaStream depthStream)
		{
			depthStream.Cleared -= OnAggregationStreamCleared;
			depthStream.End -= OnAggregationStreamEnd;
			depthStream.Flushed -= OnAggregationStreamFlushed;
		}

		/// <summary>
		/// Проверить, находятся ли все запущенные потоки в состоянии получении онлайн-данных.
		/// </summary>
		/// <returns>
		/// Находятся ли все запущенные потоки в состоянии получении онлайн-данных.
		/// </returns>
		public bool IsOnLine()
		{
			return _streamManager.Streams.All(stream => stream.IsOnline());
		}

		/// <summary>
		/// Подключиться к торговой системе.
		/// </summary>
		protected override void OnConnect()
		{
			// Для удобства проверяем, запущен ли P2MQRouter.exe?
			// Если это не делать, COM-библиотека Плазы будет выкидывать одну и ту же ошибку ('Couldn't resolve MQ service'),
			// и когда P2MQRouter.exe не запущен,
			// и когда нет связи с сервером Плазы или не ведутся торги.

			//if (_connectionPool.IsLocal && Router.Status != ServiceControllerStatus.Running)
			//    throw new InvalidOperationException("На компьютере не установлена или не запущена программа P2MQRouter.exe.");

			_transactionManager.Start();

			_connectionPool.Connect();

			this.AddInfoLog("OnConnect");
		}

		/// <summary>
		/// Отключиться от торговой системы.
		/// </summary>
		protected override void OnDisconnect()
		{
			// Иначе при выгрузке в StopStream будет срабатывать исключение.
			if (IsExportStarted)
				StopExport();

			_transactionManager.Stop();

			_connectionPool.Disconnect();

			this.AddInfoLog("OnDisconnect");
		}


		/// <summary>
		/// Запустить экспорт данных из торговой системы в программу (получение портфелей, инструментов, заявок и т.д.).
		/// </summary>
		public override void StartExport()
		{
			if (IsExportStarted)
				StopExport();

			this.AddInfoLog("Запуск экспорта.");

			_streamManager.StartExport();
			base.StartExport();
		}

		/// <summary>
		/// Остановить запущенный экспорт данных из торговой системы в программу.
		/// </summary>
		public override void StopExport()
		{
			this.AddInfoLog("Остановка экспорта.");

			if (!IsExportStarted)
			{
				this.AddInfoLog("Экспорт не был запущен.");
				return;
			}

			_streamManager.StopExport();
			base.StopExport();
		}

		private static void FillExtensionInfo(IExtendableEntity entity, PlazaRecord record)
		{
			if (record.Table.Columns.NonMandatoryColumns.IsEmpty())
				return;

			if (entity.ExtensionInfo == null)
				entity.ExtensionInfo = new Dictionary<object, object>();

			foreach (var column in record.Table.Columns.NonMandatoryColumns)
				entity.ExtensionInfo[column] = record.Get<object>(column);
		}

		private static void InitSecurityInfo(Security security, PlazaRecord record, PlazaSessionContentsDerivativeColumns metadata)
		{
			security.Exchange = Exchange.Rts;

			security.Name = record.Get<string>(metadata.Name);
			security.ShortName = record.Get<string>(metadata.Isin);
			security.MinLotSize = record.Get<int>(metadata.LotVolume);
			security.MinStepSize = record.Get<decimal>(metadata.MinStep);
			security.Decimals = record.Get<int>(metadata.RoundTo);
			security.ExpiryDate = record.Get<DateTime>(metadata.DateEndTrade);
		}

		private static void UpdateSecurityInfo(Security security, PlazaRecord record, PlazaSessionContentsDerivativeColumns metadata)
		{
			security.MinStepPrice = record.Get<decimal>(metadata.StepPrice);
			security.ClosePrice = record.Get<decimal>(metadata.LastClearingQuote);

			var isLimited = record.Get<bool>(metadata.IsLimited);
			if (isLimited)
			{
				//http://forum.rts.ru/viewtopic.asp?t=21271
				security.MaxPrice = security.ClosePrice + record.Get<decimal>(metadata.LimitUp);
				security.MinPrice = Math.Max(0, security.ClosePrice - record.Get<decimal>(metadata.LimitDown));
			}
			else
			{
				security.MaxPrice = -1;
				security.MinPrice = security.MinStepSize;
			}

			FillExtensionInfo(security, record);
		}

		private void OnSessionContentsDerivativeStreamInserted(PlazaRecord record, PlazaSessionContentsDerivativeColumns metadata, Action<Security> initHandler, Action<Security> updateHandler)
		{
			ProcessEvents(() =>
			{
				var isinId = record.Get<int>(metadata.IsinId);
				var secCode = record.Get<string>(metadata.ShortIsin);

				GetSecurity(CreateSecurityId(secCode, "RTS"), security =>
				{
					if (_isinIds.Add(isinId))
					{
						InitSecurityInfo(security, record, metadata);

						security.Code = secCode;
						initHandler(security);
					}

					UpdateSecurityInfo(security, record, metadata);
					updateHandler(security);
				}, isinId.ToString());
			});
		}

		private void OnSessionContentsFutureStreamInserted(PlazaRecord record)
		{
			var metadata = PlazaColumnRegistry.SessionContentsFuture;

			var state = (SecurityState)record.Get<int>(metadata.State);
			if (state == SecurityState.Ended)
				return; //Сессия закончена

			OnSessionContentsDerivativeStreamInserted(record, metadata, security =>
			{
				security.Class = _futuresClass;
				security.Type = SecurityTypes.Future;
				security.UnderlyingSecurityId = record.Get<string>(metadata.CodeVcb);

				//Если РЕПО или своп - добавляем в список для дальнейшей фильтрации при построении стаканов.
				if (record.Get<int>(metadata.MultilegType) != 0)
					_multilegSecurities.Add(security);
			}, security =>
			{
				security.MarginBuy = record.Get<decimal>(metadata.BuyDeposit);
				security.MarginSell = record.Get<decimal>(metadata.SellDeposit);
			});
		}

		private void OnSessionContentsOptionStreamInserted(PlazaRecord record)
		{
			var metadata = PlazaColumnRegistry.SessionContentsOption;

			OnSessionContentsDerivativeStreamInserted(record, metadata, security =>
			{
				security.Class = _optionsClass;
				security.Type = SecurityTypes.Option;
				security.ExpiryDate = record.Get<DateTime>(metadata.ExpirationStartDate);
				security.Strike = record.Get<decimal>(metadata.Strike);
				security.OptionType = record.Get<bool>(metadata.IsPut) ? OptionTypes.Put : OptionTypes.Call;

				ProcessSecurityAction(record.Get<string>(metadata.FuturesIsinId), s => security.UnderlyingSecurityId = s.Id, false);
			}, security =>
			{
				security.MarginBuy = record.Get<decimal>(metadata.BgoBuy);
			});
		}

		private void InitSecurityData(Security security, PlazaRecord record, PlazaCommonDerivativeColumns metadata)
		{
			var table = record.Table;

			if (table.Columns.Contains(metadata.BestBuy) || table.Columns.Contains(metadata.AmountToBuy))
			{
				var bestBid = new Quote
				{
					Security = security,
					OrderDirection = OrderDirections.Buy,
				};

				if (table.Columns.Contains(metadata.BestBuy))
					bestBid.Price = record.Get<decimal>(metadata.BestBuy);
				
				if (table.Columns.Contains(metadata.AmountToBuy))
					bestBid.Volume = record.Get<decimal>(metadata.AmountToBuy);

				security.BestBid = bestBid;
			}

			if (table.Columns.Contains(metadata.BestSell) || table.Columns.Contains(metadata.AmountToSell))
			{
				var bestAsk = new Quote
				{
					Security = security,
					OrderDirection = OrderDirections.Sell,
				};

				if (table.Columns.Contains(metadata.BestSell))
					bestAsk.Price = record.Get<decimal>(metadata.BestBuy);
				
				if (table.Columns.Contains(metadata.AmountToSell))
					bestAsk.Volume = record.Get<decimal>(metadata.AmountToSell);

				security.BestAsk = bestAsk;
			}

			Trade lastTrade = null;

			if (table.Columns.Contains(metadata.Price))
			{
				lastTrade = EntityFactory.CreateTrade(0);
				lastTrade.Price = record.Get<decimal>(metadata.Price);
			}

			if (table.Columns.Contains(metadata.Amount))
			{
				lastTrade = lastTrade ?? EntityFactory.CreateTrade(0);
				lastTrade.Volume = record.Get<decimal>(metadata.Amount);
			}

			if (table.Columns.Contains(metadata.DealTime))
			{
				lastTrade = lastTrade ?? EntityFactory.CreateTrade(0);
				lastTrade.Time = record.Get<DateTime>(metadata.DealTime);
			}

			if (lastTrade != null)
				security.LastTrade = lastTrade;

			if (table.Columns.Contains(metadata.MinPrice))
				security.LowPrice = record.Get<decimal>(metadata.MinPrice);

			if (table.Columns.Contains(metadata.MaxPrice))
				security.HighPrice = record.Get<decimal>(metadata.MaxPrice);

			if (table.Columns.Contains(metadata.OpeningPrice))
				security.OpenPrice = record.Get<decimal>(metadata.OpeningPrice);

			if (table.Columns.Contains(metadata.Position))
				security.OpenInterest = record.Get<decimal>(metadata.Position);

			FillExtensionInfo(security, record);
		}

		private void ProcessPlazaStream(PlazaRecord record, PlazaColumn isinId, bool ignoreIfNotExist, Action<Security> action)
		{
			ProcessEvents(() => ProcessSecurityAction(record.Get<string>(isinId), action, ignoreIfNotExist));
		}

		private void OnCommonDerivativeStreamInserted(PlazaRecord record, PlazaCommonDerivativeColumns metadata)
		{
			ProcessPlazaStream(record, metadata.IsinId, true, security => InitSecurityData(security, record, metadata));
		}

		private void OnCommonFutureStreamInserted(PlazaRecord record)
		{
			OnCommonDerivativeStreamInserted(record, PlazaColumnRegistry.CommonFuture);
		}

		private void OnCommonOptionStreamInserted(PlazaRecord record)
		{
			OnCommonDerivativeStreamInserted(record, PlazaColumnRegistry.CommonOption);
		}

		private void OnDealStreamInserted(PlazaRecord record, PlazaDealDerivativeColumns metadata)
		{
			if (OnlySystemTrades && record.Get<bool>(metadata.NoSystem))
				return;

			ProcessPlazaStream(record, metadata.IsinId, false, security =>
			{
				var trade = GetTrade(security, record.Get<long>(metadata.DealId), dealId =>
				{
					var t = EntityFactory.CreateTrade(record.Get<long>(metadata.DealId));

					t.Security = security;
					t.Price = record.Get<decimal>(metadata.Price);
					t.Volume = record.Get<decimal>(metadata.Amount);
					t.Time = record.Get<DateTime>(metadata.Moment);

					if (record.Table.Columns.Contains(metadata.BuyOrderId) && record.Table.Columns.Contains(metadata.SellOrderId))
					{
						t.OrderDirection = record.Get<long>(metadata.BuyOrderId) > record.Get<long>(metadata.SellOrderId)
						                   	? OrderDirections.Buy
						                   	: OrderDirections.Sell;
					}

					return t;
				});

				security.LastTrade = trade;

				var orderTransactionId = record.Get<int>(metadata.BuyExternalId);

				if (orderTransactionId == 0)
					orderTransactionId = record.Get<int>(metadata.SellExternalId);

				// собственная сделка
				if (orderTransactionId != 0)
				{
					AddMyTrade(security, 0, orderTransactionId, trade);
				}

				FillExtensionInfo(trade, record);
			});
		}

		private void OnDealFutureStreamInserted(PlazaRecord record)
		{
			OnDealStreamInserted(record, PlazaColumnRegistry.DealFuture);
		}

		private void OnDealOptionStreamInserted(PlazaRecord record)
		{
			OnDealStreamInserted(record, PlazaColumnRegistry.DealOption);
		}

		private void OnOrdersLogStreamInserted(PlazaStream stream, PlazaRecord record, PlazaOrdersDerivativeColumns metadata)
		{
			ProcessPlazaStream(record, metadata.IsinId, false, security => _orderBuilder.AddOrderInfo(stream, security, metadata, record));
		}

		private void OnOrdersLogFutureStreamInserted(PlazaRecord record)
		{
			OnOrdersLogStreamInserted(StreamRegistry.OrdersLogFutureStream, record, PlazaColumnRegistry.OrdersLogFuture);
		}

		private void OnOrdersLogOptionStreamInserted(PlazaRecord record)
		{
			OnOrdersLogStreamInserted(StreamRegistry.OrdersLogOptionStream, record, PlazaColumnRegistry.OrdersLogOption);
		}

		private void OnOrdersLogStreamEnd(PlazaStream stream)
		{
			ProcessEvents(() => _orderBuilder.FlushChanges(stream, (security, metadata, firstRecord, lastRecord) => GetOrder(security, firstRecord.Get<long>(metadata.OrderId), orderId =>
			{
				var transactionId = firstRecord.Get<int>(metadata.ExtId);

				//http://forum.rts.ru/viewtopic.asp?p=115933
				//10. Изменен порядок прихода записей по итогам торговой транзакции в логе заявок.
				//Ранее при постановке заявки первой шла запись о постановке заявки инициатора транзакции, 
				//вслед за которой шли записи о сделках.
				//В версии 3.8 запись о постановке заявки добавляется последней, после записей о сделках.

				var order = GetOrderByTransactionId(transactionId);

				if (order != null)
					order.Id = orderId;
				else
					order = EntityFactory.CreateOrder(orderId);

				order.Security = security;

				var restVolume = firstRecord.Get<decimal>(metadata.AmountRest);
				var initVolume = firstRecord.Get<decimal>(metadata.AmountOperation);
				var clientCode = firstRecord.Get<string>(metadata.ClientCode);

				order.Price = firstRecord.Get<decimal>(metadata.Price);
				order.Volume = initVolume;
				order.Balance = restVolume;
				order.Direction = firstRecord.Get<int>(metadata.Direction) == 1 ? OrderDirections.Buy : OrderDirections.Sell;
				if (!clientCode.IsEmpty())
					order.Portfolio = GetPortfolio(clientCode);
				order.Comment = firstRecord.Get<string>(metadata.Comment);
				order.TransactionId = transactionId;
				order.Time = firstRecord.Get<DateTime>(metadata.Moment);
				order.State = restVolume == 0 ? OrderStates.Done : OrderStates.Active;

				var status = firstRecord.Get<int>(metadata.Status);

				if ((status & 0x01) == 0x01)
					order.ExecutionCondition = OrderExecutionConditions.PutInQueue;
				else if ((status & 0x02) == 0x02)
					order.ExecutionCondition = OrderExecutionConditions.CancelBalance;

				return order;
			},
			order =>
			{
				var action = lastRecord.Get<int>(metadata.Action);
				switch (action)
				{
					case 0: // удалена
						order.CancelTime = lastRecord.Get<DateTime>(metadata.Moment);
						order.Balance = lastRecord.Get<int>(metadata.AmountOperation);
						order.State = OrderStates.Done;
						break;
					case 1: // добавлена
						break;
					case 2: // исполнена
						order.Balance = lastRecord.Get<decimal>(metadata.AmountRest);
						if (order.Balance == 0)
							order.State = OrderStates.Done;
						break;
					default:
						throw new InvalidOperationException("Неизвестное действие заявки {0}.".Put(action));
				}

				FillExtensionInfo(order, lastRecord);
			})));
		}

		private void OnOrdersLogStreamFlushed(PlazaStream stream)
		{
			_orderBuilder.ClearChanges(stream);
		}

		private void OnPortfoliosStreamInserted(PlazaRecord record)
		{
			var metadata = PlazaColumnRegistry.Portfolios;

			ProcessEvents(() => GetPortfolio(record.Get<string>(metadata.ClientCode), portfolio =>
			{
				portfolio.Name = record.Get<string>(metadata.ClientCode);
				portfolio.BeginAmount = record.Get<decimal>(metadata.MoneyOld);
				portfolio.CurrentAmount = record.Get<decimal>(metadata.MoneyAmount);
				portfolio.Commission = record.Get<decimal>(metadata.Fee);
				portfolio.Leverage = record.Get<decimal>(metadata.CoeffGo);
				portfolio.Exchange = Exchange.Rts;

				FillExtensionInfo(portfolio, record);
			}));
		}

		private void OnPositionsStreamInserted(PlazaRecord record)
		{
			var metadata = PlazaColumnRegistry.Positions;

			ProcessPlazaStream(record, metadata.IsinId, false, security =>
				GetPosition(GetPortfolio(record.Get<string>(metadata.ClientCode)), security,
					position =>
					{
						position.BeginValue = record.Get<decimal>(metadata.OpenQty);
						position.CurrentValue = record.Get<decimal>(metadata.Pos);
						FillExtensionInfo(position, record);
					}));
		}

		private void OnAggregation5OptionStreamInserted(PlazaRecord record)
		{
			OnAggregationStreamInserted(StreamRegistry.Aggregation5OptionStream, record, PlazaColumnRegistry.Aggregation5Option);
		}

		private void OnAggregation5FutureStreamInserted(PlazaRecord record)
		{
			OnAggregationStreamInserted(StreamRegistry.Aggregation5FutureStream, record, PlazaColumnRegistry.Aggregation5Future);
		}

		private void OnAggregation20OptionStreamInserted(PlazaRecord record)
		{
			OnAggregationStreamInserted(StreamRegistry.Aggregation20OptionStream, record, PlazaColumnRegistry.Aggregation20Option);
		}

		private void OnAggregation20FutureStreamInserted(PlazaRecord record)
		{
			OnAggregationStreamInserted(StreamRegistry.Aggregation20FutureStream, record, PlazaColumnRegistry.Aggregation20Future);
		}

		private void OnAggregation50OptionStreamInserted(PlazaRecord record)
		{
			OnAggregationStreamInserted(StreamRegistry.Aggregation50OptionStream, record, PlazaColumnRegistry.Aggregation50Option);
		}

		private void OnAggregation50FutureStreamInserted(PlazaRecord record)
		{
			OnAggregationStreamInserted(StreamRegistry.Aggregation50FutureStream, record, PlazaColumnRegistry.Aggregation50Future);
		}

		private void OnAggregationStreamFlushed(PlazaStream stream)
		{
			_depthBuilder.ClearQuotes(stream);
		}

		private void OnAggregationStreamInserted(PlazaStream stream, PlazaRecord record, PlazaAggregationColumns metadata)
		{
			//Trace.WriteLine("OnADI rec "+System.Threading.Thread.CurrentThread.ManagedThreadId);

			ProcessPlazaStream(record, metadata.IsinId, true, security =>
			{
				//Стаканы для свопов и РЕПО строятся иначе - может быть Bid>=Ask. Для свопов стаканы строятся по процентам.
				if (!_multilegSecurities.Contains(security))
					_depthBuilder.UpdateQuote(stream, security, record, metadata);
			});
		}

		private void OnAggregationStreamCleared(PlazaStream stream, long revision)
		{
			_depthBuilder.RemoveQuotesByRev(stream, revision);
		}

		private void OnAggregationStreamEnd(PlazaStream stream)
		{
			RaiseQuotesChanged(_depthBuilder.FlushChanges(stream));
		}

		private void OnSessionFutureStreamInserted(PlazaRecord record)
		{
			var metadata = PlazaColumnRegistry.SessionFuture;

			SessionState = (SessionState)record.Get<int>(metadata.State);

			// http://forum.rts.ru/viewtopic.asp?t=20203
			// Снимаем активные заявки с датой истечения сегодня
			if (SessionState == SessionState.Ended)
			{
				foreach (var order in Orders.Where(o => o.State == OrderStates.Active && o.ExpiryDate == null))
				{
					order.CancelTime = MarketTime;
					order.State = OrderStates.Done;

					RaiseOrderChanged(order);
				}
			}
		}

		private void OnVolatilityStreamInserted(PlazaRecord record)
		{
			var metadata = PlazaColumnRegistry.Volatility;

			ProcessPlazaStream(record, metadata.IsinId, true, security =>
			{
				security.Volatility = record.Get<decimal>(metadata.Volatility);
				security.TheorPrice = record.Get<decimal>(metadata.TheorPrice);
			});
		}

		/// <summary>
		/// Зарегистрировать заявку на бирже.
		/// </summary>
		/// <param name="order">Заявка, содержащая информацию для регистрации.</param>
		protected override void OnRegisterOrder(Order order)
		{
			if (order.Type == OrderTypes.Conditional)
				throw new NotSupportedException("Условные заявки не поддерживаются протоколом Plaza2.");

			var transaction = _transactionManager.Factory.CreateRegister(order);

			transaction
				.SetPrice(order.Price)
				.SetVolume((int)order.Volume)
				.SetSecurity(order.Security)
				.SetClientCode(order.Portfolio.ToClientCode())
				.SetDirection(order.Direction)
				.SetExecutionCondition(order.ExecutionCondition)
				.SetExtId((int)order.TransactionId);

			if (!order.Comment.IsEmpty())
				transaction.SetComment(order.Comment);

			if (order.ExpiryDate != null)
				transaction.SetExpiryDate(order.ExpiryDate.Value);

			_transactionManager.SendTransaction(transaction);
		}

		/// <summary>
		/// Отменить заявку на бирже.
		/// </summary>
		/// <param name="order">Заявка, которую нужно отменять.</param>
		protected override void OnCancelOrder(Order order)
		{
			if (order.Id == 0)
			{
				//Отменяем заявку по внешнему номеру
				var transaction = CreateCancelGroupTransaction(order.Security, order.Direction, order.Portfolio, order.TransactionId, order);
				_transactionManager.SendTransaction(transaction);
			}
			else
			{
				var transaction = _transactionManager.Factory.CreateCancel(order);
				transaction.SetOrderId(order.Id);
				_transactionManager.SendTransaction(transaction);
			}
		}

		/// <summary>
		/// Перерегистрировать заявку на бирже.
		/// </summary>
		/// <param name="oldOrder">Заявка, которую нужно снять.</param>
		/// <param name="newOrder">Новая заявка, которую нужно зарегистрировать.</param>
		protected override void OnReRegisterOrder(Order oldOrder, Order newOrder)
		{
			InitNewOrder(newOrder);

			var transaction = _transactionManager.Factory.CreateReRegister(oldOrder, newOrder);

			transaction
				.SetMoveMode(oldOrder.Volume == newOrder.Volume)
				.SetFirstOrderId(oldOrder.Id)
				.SetFirstVolume((int)newOrder.Volume)
				.SetFirstPrice(newOrder.Price)
				.SetFirstExtId((int)newOrder.TransactionId);

			_transactionManager.SendTransaction(transaction);
		}

		/// <summary>
		/// Перерегистрировать пару заявок на бирже.
		/// </summary>
		/// <param name="oldOrder1">Первая заявка, которую нужно снять.</param>
		/// <param name="newOrder1">Первая новая заявка, которую нужно зарегистрировать.</param>
		/// <param name="oldOrder2">Вторая заявка, которую нужно снять.</param>
		/// <param name="newOrder2">Вторая новая заявка, которую нужно зарегистрировать.</param>
		protected override void OnReRegisterOrderPair(Order oldOrder1, Order newOrder1, Order oldOrder2, Order newOrder2)
		{

			InitNewOrder(newOrder1);
			InitNewOrder(newOrder2);

			var transaction = _transactionManager.Factory.CreateReRegisterPair(oldOrder1, newOrder1, oldOrder2, newOrder2);

			transaction
				.SetMoveMode(oldOrder1.Volume == newOrder1.Volume && oldOrder2.Volume == newOrder2.Volume)
				.SetFirstOrderId(oldOrder1.Id)
				.SetFirstVolume((int)newOrder1.Volume)
				.SetFirstPrice(newOrder1.Price)
				.SetFirstExtId((int)newOrder1.TransactionId)

				.SetSecondOrderId(oldOrder2.Id)
				.SetSecondVolume((int)newOrder2.Volume)
				.SetSecondPrice(newOrder2.Price)
				.SetSecondExtId((int)newOrder2.TransactionId);

			_transactionManager.SendTransaction(transaction);
		}

		/// <summary>
		/// Отменить группу заявок на бирже по фильтру.
		/// </summary>
		/// <param name="isStopOrder">True, если нужно отменить только стоп-заявки, false - если только обычный и null - если оба типа.</param>
		/// <param name="portfolio">Портфель. Если значение равно null, то портфель не попадает в фильтр снятия заявок.</param>
		/// <param name="direction">Направление заявки. Если значение равно null, то направление не попадает в фильтр снятия заявок.</param>
		/// <param name="classCode">Код класса. Если переданная строка пустая, то код не попадает в фильтр снятия заявок.</param>
		/// <param name="security">Инструмент. Если значение равно null, то инструмент не попадает в фильтр снятия заявок.</param>
		public override void CancelOrders(bool? isStopOrder = null, Portfolio portfolio = null, OrderDirections? direction = null, string classCode = null, Security security = null)
		{
			if (security == null)
				base.CancelOrders(isStopOrder, portfolio, direction, classCode, null);
			else
			{
				//TODO: передавать заявки на отмену при создании транзакции
				//var orders = Orders.Where(order =>
				//    {
				//        if (order.State != OrderStates.Active)
				//        {
				//            return false;
				//        }

				//        if (isStopOrder != null)
				//        {
				//            if (isStopOrder.Value && order.Type != OrderTypes.Conditional)
				//            {
				//                return false;
				//            }
				//            else if (order.Type != OrderTypes.Limit)
				//            {
				//                return false;
				//            }
				//        }

				//        if (portfolio != null && order.Portfolio != portfolio)
				//        {
				//            return false;
				//        }

				//        if (direction != null && order.Direction != direction)
				//        {
				//            return false;
				//        }

				//        if (!classCode.IsEmpty() && order.Security.Class != classCode)
				//        {
				//            return false;
				//        }

				//        if (security != null && order.Security == security)
				//        {
				//            return false;
				//        }

				//        return true;
				//    }
				//);

				//var transaction = CreateCancelGroupTransaction(security, direction, portfolio, 0, orders.ToArray());
				var transaction = CreateCancelGroupTransaction(security, direction, portfolio, 0);
				_transactionManager.SendTransaction(transaction);
			}
		}

		private Transaction CreateCancelGroupTransaction(Security security, OrderDirections? direction, Portfolio portfolio, long ext_id, params Order[] orders)
		{
			var transaction = _transactionManager.Factory.CreateCancelGroup(security.Type, orders);

			transaction
				.SetSecurity(security)
				.SetDirection(direction)
				.SetOrderType(null)
				.SetExtId((int)ext_id);

			if (portfolio != null)
				transaction.SetClientCode(portfolio.ToClientCode());

			return transaction;
		}

		private void ProcessReply(Transaction replyTransaction)
		{
			if (!replyTransaction.CheckOrdersLength())
			{
				throw new ArgumentException("Неверное число заявок ({0}) в транзакции {1}".Put(
					replyTransaction.Orders.Length, replyTransaction.Message), "replyTransaction");
			}

			ProcessEvents(() =>
			{
				if (replyTransaction.ErrorInfo != null)
				{
					foreach (var order in replyTransaction.Orders)
					{
						RaiseOrderFailed(order, replyTransaction.ErrorInfo);
					}
				}
				else
				{
					switch (replyTransaction.OriginTransaction.Message)
					{
						case Messages.FutureRegister:
						case Messages.OptionRegister:
						{
							var order = replyTransaction.Orders[0];
							AddNewOrderByIdIfNeeded(replyTransaction, "order_id", order);
							break;
						}
						case Messages.FutureReRegister:
						case Messages.OptionReRegister:
						{
							switch (replyTransaction.Orders.Length)
							{
								case 4: //Перерегистрация пары заявок
								{
									var oldOrder1 = replyTransaction.Orders[0];
									CancelOrderIfNeeded(oldOrder1);

									var newOrder1 = replyTransaction.Orders[1];
									AddNewOrderByIdIfNeeded(replyTransaction, "order_id1", newOrder1);

									var oldOrder2 = replyTransaction.Orders[2];
									CancelOrderIfNeeded(oldOrder2);

									var newOrder2 = replyTransaction.Orders[3];
									AddNewOrderByIdIfNeeded(replyTransaction, "order_id2", newOrder2);
									break;
								}
								case 2: //Перерегистрация одной заявки
								{
									var oldOrder = replyTransaction.Orders[0];
									CancelOrderIfNeeded(oldOrder);

									var newOrder = replyTransaction.Orders[1];
									AddNewOrderByIdIfNeeded(replyTransaction, "order_id1", newOrder);

									break;
								}
							}

							break;
						}
						case Messages.FutureCancel:
						case Messages.OptionCancel:
						{
							var order = replyTransaction.Orders[0];
							CancelOrderIfNeeded(order, (int)replyTransaction.GetField("amount"));
							break;
						}
						case Messages.FutureCancelGroup:
						case Messages.OptionCancelGroup:
						{
							var ordersCount = (int)replyTransaction.GetField("num_orders");

							if (ordersCount == replyTransaction.Orders.Length)
							{
								foreach (var order in replyTransaction.Orders)
									CancelOrderIfNeeded(order);
							}
							else if (ordersCount == 0)
							{
								foreach (var order in replyTransaction.Orders)
									RaiseOrderFailed(order, new Exception("Не удалось снять заявку {0}".Put(order.TransactionId)));
							}

							break;
						}
						default:
							throw new ArgumentOutOfRangeException();
					}
				}
			});
		}

		private void AddNewOrderByIdIfNeeded(Transaction transaction, string idFieldName, Order order)
		{
			if (order.Id != 0)
				return;

			var orderId = (long)transaction.GetField(idFieldName);
			order.Id = orderId;

			if (order.Id > 0)
			{
				GetOrder(order.Security, orderId, key =>
				{
					order.Time = MarketTime;

					if (order.State == OrderStates.None)
						order.State = OrderStates.Active;

					return order;
				}, o => { });
			}
			else
			{
				var exception = new InvalidOperationException("Не удалось зарегистрировать заявку {0}.".Put(order.TransactionId));
				RaiseOrderFailed(order, exception);
			}
		}

		private void CancelOrderIfNeeded(Order order, int? newBalance = null)
		{
			//Проверка на OrderStates.None нужна, т.к. можем снять заявку по ext_id,
			//при этом получить вначале ответ по Cancel, затем лишь по Register.
			if (order.State == OrderStates.None || order.State == OrderStates.Active)
			{
				if (newBalance != null)
				{
					order.Balance = newBalance.Value;
				}

				order.CancelTime = MarketTime;
				order.State = OrderStates.Done;
				RaiseOrderChanged(order);
			}
		}

		/// <summary>
		/// Начать получать котировки (стакан) по инструменту.
		/// Значение котировок можно получить через метод <see cref="M:StockSharp.BusinessEntities.ITrader.GetMarketDepth(StockSharp.BusinessEntities.Security)"/>.
		/// </summary>
		/// <param name="security">Инструмент, по которому необходимо начать получать котировки.</param>
		public override void RegisterQuotes(Security security)
		{
		}

		/// <summary>
		/// Остановить получение котировок по инструменту.
		/// </summary>
		/// <param name="security">Инструмент, по которому необходимо остановить получение котировок.</param>
		public override void UnRegisterQuotes(Security security)
		{
		}

		/// <summary>
		/// Начать получать сделки (тиковые данные) по инструменту. Новые сделки будут приходить через событие <see cref="E:StockSharp.BusinessEntities.ITrader.NewTrades"/>.
		/// </summary>
		/// <param name="security">Инструмент, по которому необходимо начать получать сделки.</param>
		public override void RegisterTrades(Security security)
		{
		}

		/// <summary>
		/// Остановить получение сделок (тиковые данные) по инструменту.
		/// </summary>
		/// <param name="security">Инструмент, по которому необходимо остановить получение сделок.</param>
		public override void UnRegisterTrades(Security security)
		{
		}

		/// <summary>
		/// Начать получать новую информацию по портфелю.
		/// </summary>
		/// <param name="portfolio">Портфель, по которому необходимо начать получать новую информацию.</param>
		public override void RegisterPortfolio(Portfolio portfolio)
		{
		}

		/// <summary>
		/// Остановить получение новой информации по портфелю.
		/// </summary>
		/// <param name="portfolio">Портфель, по которому необходимо остановить получение новой информации.</param>
		public override void UnRegisterPortfolio(Portfolio portfolio)
		{
		}

		/// <summary>
		/// Начать получать новую информацию (например, <see cref="P:StockSharp.BusinessEntities.Security.LastTrade"/> или <see cref="P:StockSharp.BusinessEntities.Security.BestBid"/>) по инструменту.
		/// </summary>
		/// <param name="security">Инструмент, по которому необходимо начать получать новую информацию.</param>
		public override void RegisterSecurity(Security security)
		{
		}

		/// <summary>
		/// Остановить получение новой информации.
		/// </summary>
		/// <param name="security">Инструмент, по которому необходимо остановить получение новой информации.</param>
		public override void UnRegisterSecurity(Security security)
		{
		}

		/// <summary>
		/// Загрузить настройки.
		/// </summary>
		/// <param name="settings">Хранилище настроек.</param>
		public override void Load(SettingsStorage settings)
		{
			base.Load(settings);

			Address = settings.GetValue<IPEndPoint>("Address");
			Login = settings.GetValue<string>("Login");
			Password = settings.GetValue<string>("Password");
			AppName = settings.GetValue<string>("AppName");
			UseLocalProtocol = settings.GetValue<bool>("UseLocalProtocol");
			StreamTimeOut = settings.GetValue<TimeSpan>("StreamTimeOut");
			StreamConfigPath = settings.GetValue<string>("StreamConfigPath");
			TransactionManager.IsAsync = settings.GetValue<bool>("IsAsync");
			OnlySystemTrades = settings.GetValue<bool>("OnlySystemTrades");

			var defaultFutureDepthStream = settings.GetValue<string>("DefaultFutureDepthStream");
			DefaultFutureDepthStream = StreamRegistry.AggregationFutureStreams.First(s => s.Table.Id.ToString() == defaultFutureDepthStream);

			var defaultOptionDepthStream = settings.GetValue<string>("DefaultOptionDepthStream");
			DefaultOptionDepthStream = StreamRegistry.AggregationOptionStreams.First(s => s.Table.Id.ToString() == defaultOptionDepthStream);
		}

		/// <summary>
		/// Сохранить настройки.
		/// </summary>
		/// <param name="settings">Хранилище настроек.</param>
		public override void Save(SettingsStorage settings)
		{
			base.Save(settings);

			settings.SetValue("Address", Address.To<string>());
			settings.SetValue("Login", Login);
			settings.SetValue("Password", Password);
			settings.SetValue("AppName", AppName);
			settings.SetValue("UseLocalProtocol", UseLocalProtocol);
			settings.SetValue("StreamTimeOut", StreamTimeOut);
			settings.SetValue("StreamConfigPath", StreamConfigPath);
			settings.SetValue("IsAsync", TransactionManager.IsAsync);
			settings.SetValue("OnlySystemTrades", OnlySystemTrades);

			settings.SetValue("DefaultFutureDepthStream", DefaultFutureDepthStream.Table.Id.ToString());
			settings.SetValue("DefaultOptionDepthStream", DefaultOptionDepthStream.Table.Id.ToString());
		}

		/// <summary>
		/// Освободить занятые ресурсы.
		/// </summary>
		protected override void DisposeManaged()
		{
			UnSubscribeStreamRegistryEvents(StreamRegistry);
			UnSubscribeDepthStream(DefaultFutureDepthStream);
			UnSubscribeDepthStream(DefaultOptionDepthStream);

			StopExport();

			_streamManager.Dispose();

			_transactionManager.ProcessResponse -= ProcessReply;

			_connectionPool.Dispose();
		}
	}
}
