﻿namespace StockSharp.Plaza
{
	using System;
	using System.Collections.Generic;
	using System.Linq;

	using Ecng.Collections;
	using Ecng.Common;

	using StockSharp.Algo;
	using StockSharp.Plaza.Wrappers;
	
	/// <summary>
	/// Менеджер транзакций.
	/// </summary>
	public class TransactionManager
	{
		private readonly MessageDispatcher _messageDispatcher;
		private readonly PlazaConnectionPool _connectionPool;

		private readonly SynchronizedMultiDictionary<long, Transaction> _transactionsByTransactionIds = new SynchronizedMultiDictionary<long, Transaction>();

		private readonly SynchronizedQueue<Transaction> _transactionQueue = new SynchronizedQueue<Transaction>();
		private readonly SynchronizedList<PlazaConnectionToken> _runningConnections = new SynchronizedList<PlazaConnectionToken>();

		internal TransactionManager(PlazaConnectionPool connectionPool, TransactionIdGenerator idGenerator)
		{
			if (connectionPool == null)
				throw new ArgumentNullException("connectionPool");

			_connectionPool = connectionPool;

			Factory = new TransactionFactory(connectionPool, idGenerator);
            TransactionTimeOut = TimeSpan.FromSeconds(5);

			_messageDispatcher = new MessageDispatcher((msg, errorCode, transactionId) =>
			{
				var transaction = GetTransactions(transactionId).Single();
				if (errorCode != 0)
				{
					//http://forum.rts.ru/viewtopic.asp?p=118911
					TransactionFailed.SafeInvoke(transaction, errorCode);
				}
				else
				{
					ProcessReply(transaction, msg);
				}
			});

			IsAsync = true;
			
			// Необходимо создать хотя бы одно транзакционное подключение
			ThreadCount = 1;
		}

		/// <summary>
		/// Использовать ли асинхронный режим при отправке транзакций.
		/// </summary>
		/// <remarks>
		/// Значение по умолчанию равно true.
		/// </remarks>
		public bool IsAsync { get; set; }

		/// <summary>
		/// Фабрика транзакций.
		/// </summary>
		public TransactionFactory Factory { get; private set; }

		/// <summary>
		/// Время, в течении которого ожидается ответ для транзакции.
		/// </summary>
		/// <remarks>
		/// По-умолчанию равно 5 секундам.
		/// </remarks>
		public TimeSpan TransactionTimeOut { get; set; }

		/// <summary>
		/// Обработать отправляемую транзакцию до того, как она будет отослана на сервер Plaza.
		/// </summary>
		public event Action<Transaction> ProcessRequest;

		/// <summary>
		/// Обработать полученную транзакцию до того, как она будет обработана <see cref="PlazaTrader"/>.
		/// </summary>
		public event Action<Transaction> ProcessResponse;

		/// <summary>
		/// Ответ на посланную транзакцию не успел придти за <see cref="TransactionTimeOut"/>.
		/// </summary>
		public event Action<Transaction, uint> TransactionFailed;

		/// <summary>
		/// Отправить транзакцию на сервер Plaza.
		/// </summary>
		/// <param name="transaction">Транзакция.</param>
		public void SendTransaction(Transaction transaction)
		{
			if (transaction == null)
				throw new ArgumentNullException("transaction");

			ProcessRequest.SafeInvoke(transaction);

			AddTransaction(transaction);
			_transactionQueue.Enqueue(transaction);
		}

		private int _threadCount;

		/// <summary>
		/// Число транзакционных потоков.
		/// </summary>
		/// <remarks>
		/// Значение по умолчанию равно 1.
		/// </remarks>
		public int ThreadCount
		{
			get { return _threadCount; }
			set
			{
				if (value < 1)
					throw new ArgumentOutOfRangeException("value", value, @"Число транзакционных потоков не может быть меньше 1.");

				if (_runningConnections.Count > 0)
					throw new InvalidOperationException("Нельзя менять количество транзакционных потоков после из запуска.");

				if (value == ThreadCount)
					return;

				_threadCount = value;
			}
		}

		internal void Start()
		{
			if (_runningConnections.Count > 0)
				throw new InvalidOperationException("Транзакционные потоки не были остановлены с предыдущего запуска.");

			for (var i = 0; i < ThreadCount; i++)
			{
				var token = _connectionPool.RunConnection("Plaza. Transaction thread #{0}".Put(i + 1), connection =>
				{
					while (true)
					{
						var newTransaction = _transactionQueue.TryDequeue();
						if (newTransaction == null)
							break;

						var timeOut = (uint)TransactionTimeOut.TotalMilliseconds;
						var plazaMessage = newTransaction.PlazaMessage;

						if (IsAsync)
						{
							plazaMessage.SendAsync(connection, timeOut, _messageDispatcher, newTransaction.Id);
						}
						else
						{
							var reply = plazaMessage.Send(connection, timeOut);
							ProcessReply(newTransaction, reply);
						}
					}

					return true;
				});

				_runningConnections.Add(token);
			}
		}

		internal void Stop()
		{
			_runningConnections.ForEach(token => _connectionPool.StopConnection(token));
			_runningConnections.Clear();
		}

		/// <summary>
		/// Получить транзакции (запрос и ответ) по номеру.
		/// </summary>
		/// <param name="transactionId">Номер транзакции.</param>
		/// <returns>Транзакции (запрос и ответ).</returns>
		public IEnumerable<Transaction> GetTransactions(long transactionId)
		{
			return _transactionsByTransactionIds.TryGetValue(transactionId);
		}

		private void AddTransaction(Transaction transaction)
		{
			if (transaction == null)
				throw new ArgumentNullException("transaction");

			_transactionsByTransactionIds.Add(transaction.Id, transaction);
		}

		private void ProcessReply(Transaction originMessage, IMessageWrapper reply)
		{
			var replyTransaction = new Transaction(originMessage, reply);
			AddTransaction(replyTransaction);

			PlazaException errorInfo;

			if (replyTransaction.GetCategory() == TransactionFactory.MessagesCategory)
			{
				var code = replyTransaction.GetReplyCode();
				errorInfo = code == 0 ? null : new PlazaException(code, replyTransaction.GetMessage());
			}
			else
				errorInfo = new PlazaException(replyTransaction.GetErrorCode(), "Неизвестная ошибка.");

			replyTransaction.ErrorInfo = errorInfo;

			ProcessResponse.SafeInvoke(replyTransaction);
		}
	}
}